# -*- coding: utf-8 -*-
__version__ = "0.2.33"
__author__ = "Domingo E. Savoretti"
__email__ = "esavoretti@gmail.com"
__license__ = "MIT"
__url__ = "https://github.com/sandy98/oxi"
__description__ = "ASGI full stack: Web Server, Web App and a ton of goodies."
__copyright__ = "Copyright (c) 2025 Domingo E. Savoretti"
__oxi_port__ = 8086
__oxi_host__ = "0.0.0.0"

__all__ = [
    "utils",
    "mp4parser",
    "template",
    "websocket",
    "app",
    "server",
    "config",
]